/*
 * Copyright (C) 2013 Brockmann Consult GmbH (info@brockmann-consult.de)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option)
 * any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see http://www.gnu.org/licenses/
 */

package com.bc.ceres.binding.dom;

import com.bc.ceres.binding.ConversionException;
import com.bc.ceres.binding.Converter;
import com.bc.ceres.binding.ConverterRegistry;
import com.bc.ceres.binding.PropertyDescriptor;
import com.bc.ceres.binding.PropertyDescriptorFactory;
import com.bc.ceres.binding.ValidationException;
import com.thoughtworks.xstream.io.copy.HierarchicalStreamCopier;
import com.thoughtworks.xstream.io.xml.XppDomWriter;
import com.thoughtworks.xstream.io.xml.XppReader;
import com.thoughtworks.xstream.io.xml.xppdom.XppDom;
import junit.framework.AssertionFailedError;
import junit.framework.TestCase;

import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.Arc2D;
import java.awt.geom.Line2D;
import java.io.File;
import java.io.StringReader;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

public class DefaultDomConverterTest extends TestCase {

    private static final PropertyDescriptorFactory VALUE_DESCRIPTOR_FACTORY = new PropertyDescriptorFactory() {
        @Override
        public PropertyDescriptor createValueDescriptor(java.lang.reflect.Field field) {
            final PropertyDescriptor descriptor = new PropertyDescriptor(field.getName(), field.getType());
            final X xAnnotation = field.getAnnotation(X.class);
            if (xAnnotation != null) {
                descriptor.setAlias(xAnnotation.alias());
                descriptor.setItemAlias(xAnnotation.componentAlias());
                descriptor.setItemsInlined(xAnnotation.inlined());
                if (xAnnotation.defaultValue() != null && !xAnnotation.defaultValue().isEmpty()) {
                    try {
                        descriptor.setDefaultConverter();
                        descriptor.setDefaultValue(descriptor.getConverter().parse(xAnnotation.defaultValue()));
                    } catch (ConversionException e) {
                        throw new IllegalStateException("Failed to convert default value.", e);
                    }
                }
                if (xAnnotation.domConverter() != DomConverter.class) {
                    DomConverter domConverter;
                    try {
                        domConverter = xAnnotation.domConverter().newInstance();
                    } catch (Throwable t) {
                        throw new IllegalStateException("Failed to create domConverter.", t);
                    }
                    descriptor.setDomConverter(domConverter);
                }
            }
            return descriptor;
        }
    };

    public void testUnknownElement() throws Exception {
        final String xml = ""
                           + "<parameters>"
                           + "  <kornField>42</kornField>"
                           + "</parameters>";
        final XppDom dom = createDom(xml);
        final SimplePojo value = new SimplePojo();
        try {
            convertDomToValue(dom, value);
            fail("ConversionException expected");
        } catch (ConversionException ignored) {
        }
    }

    public void testDomToSimplePojo() throws Exception {
        final String xml = ""
                           + "<parameters>"
                           + "  <intField>42</intField>"
                           + "  <stringField>a string</stringField>"
                           + "  <doubleArrayField>1.2, 4.5, -0.034</doubleArrayField>"
                           + "  <fileField>C:/data/MER.N1</fileField>"
                           + "</parameters>";
        final XppDom dom = createDom(xml);
        final SimplePojo value = new SimplePojo();
        assertEquals(0, value.intField);
        assertEquals(null, value.stringField);
        assertNull(value.doubleArrayField);
        assertNull(value.fileField);
        convertDomToValue(dom, value);
        assertEquals(42, value.intField);
        assertEquals("a string", value.stringField);
        assertEquals(new File("C:/data/MER.N1"), value.fileField);
        assertNotNull(value.doubleArrayField);
        assertEquals(3, value.doubleArrayField.length);
        assertEquals(1.2, value.doubleArrayField[0], 1.0e-10);
        assertEquals(4.5, value.doubleArrayField[1], 1.0e-10);
        assertEquals(-0.034, value.doubleArrayField[2], 1.0e-10);
    }

    public void testSimplePojoToDom() throws ValidationException, ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <fileField>" + new File("C:/data/dat.ini") + "</fileField>"
                                   + "  <intField>43</intField>"
                                   + "  <stringField>This is a test.</stringField>"
                                   + "  <doubleArrayField>0.1,0.2,-0.4</doubleArrayField>"
                                   + "</parameters>";
        final XppDom dom = new XppDom("parameters");
        final SimplePojo value = new SimplePojo();
        value.intField = 43;
        value.stringField = "This is a test.";
        value.doubleArrayField = new double[]{0.1, 0.2, -0.4};
        value.fileField = new File("C:/data/dat.ini");
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testSimplePojoWithNullArrayToDom() throws ValidationException, ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <fileField>" + new File("C:/data/dat.ini") + "</fileField>"
                                   + "  <intField>43</intField>"
                                   + "  <stringField>This is a test.</stringField>"
                                   + "  <doubleArrayField/>"
                                   + "</parameters>";
        final XppDom dom = new XppDom("parameters");
        final SimplePojo value = new SimplePojo();
        value.intField = 43;
        value.stringField = "This is a test.";
        value.doubleArrayField = null;
        value.fileField = new File("C:/data/dat.ini");
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testSimplePojoWithNullObjectToDom() throws ValidationException, ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <fileField/>"
                                   + "  <intField>43</intField>"
                                   + "  <stringField>This is a test.</stringField>"
                                   + "  <doubleArrayField>0.1,0.2,-0.4</doubleArrayField>"
                                   + "</parameters>";
        final XppDom dom = new XppDom("parameters");
        final SimplePojo value = new SimplePojo();
        value.intField = 43;
        value.stringField = "This is a test.";
        value.doubleArrayField = new double[]{0.1, 0.2, -0.4};
        value.fileField = null;
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testArrayWithDomConverterPojoToDom() throws ValidationException, ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <allies>"
                                   + "  <member><name>bibo</name></member>"
                                   + "  <member><name>mimi</name></member>"
                                   + "  </allies>"
                                   + "</parameters>";
        final XppDom dom = new XppDom("parameters");
        final ArrayWithDomConverterPojo value = new ArrayWithDomConverterPojo();
        value.allies = new Member[2];
        value.allies[0] = new Member("bibo");
        value.allies[1] = new Member("mimi");
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testArrayWithDomConverterPojoToDom_NullValue() throws ValidationException, ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "</parameters>";
        final XppDom dom = new XppDom("parameters");
        final ArrayWithDomConverterPojo value = new ArrayWithDomConverterPojo();
        value.allies = null;
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }


    public void testWeirdPojoToDom() throws ValidationException, ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <finalWeird/>"
                                   + "  <weird/>"
                                   + "  <name>ernie</name>"
                                   + "</parameters>";
        final XppDom dom = new XppDom("parameters");
        final WeirdPojo value = new WeirdPojo();
        value.weird = new Weird();
        value.name = "ernie";
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testDomToArrayPojo() throws ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <fileField>" + new File("C:/data/dat.ini") + "</fileField>"
                                   + "  <intField>43</intField>"
                                   + "  <stringField>This is a test.</stringField>"
                                   + "  <doubleArrayField>0.1,0.2,-0.4</doubleArrayField>"
                                   + "</parameters>";
        final XppDom dom = new XppDom("parameters");
        final SimplePojo value = new SimplePojo();
        value.intField = 43;
        value.stringField = "This is a test.";
        value.doubleArrayField = new double[]{0.1, 0.2, -0.4};
        value.fileField = new File("C:/data/dat.ini");
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testArrayPojoToDom() throws ConversionException {
        ConverterRegistry.getInstance().setConverter(Member.class, new Converter<Member>() {
            @Override
            public Class<? extends Member> getValueType() {
                return Member.class;
            }

            @Override
            public Member parse(String text) throws ConversionException {
                return new Member(text);
            }

            @Override
            public String format(Member field) {
                return field.name;
            }
        });

        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <prince>bert</prince>"
                                   + "  <allies>bibo,mimi</allies>"
                                   + "</parameters>";

        final XppDom dom = new XppDom("parameters");
        final ArrayPojo arrayPojo = new ArrayPojo();
        arrayPojo.prince = new Member("bert");
        arrayPojo.allies = new Member[2];
        arrayPojo.allies[0] = new Member("bibo");
        arrayPojo.allies[1] = new Member("mimi");

        convertValueToDom(arrayPojo, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testDomToAnnotatedPojo() throws Exception {

        final String xml = ""
                           + "<parameters>"
                           + "  <targetBand>result</targetBand>"
                           + "  <targetBands>"
                           + "    <band>a</band>"
                           + "    <band>b</band>"
                           + "  </targetBands>"
                           + "</parameters>";
        final XppDom dom = createDom(xml);
        final AnnotatedPojo value = new AnnotatedPojo();
        assertNull(value.targetBandName);
        assertNull(value.targetBandNames);
        convertDomToValue(dom, value);
        assertEquals("result", value.targetBandName);
        assertNotNull(value.targetBandNames);
        assertEquals(2, value.targetBandNames.length);
        assertEquals("a", value.targetBandNames[0]);
        assertEquals("b", value.targetBandNames[1]);
    }

    public void testAnnotatedPojoToDom() throws ValidationException, ConversionException {

        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <targetBand>radiance_13</targetBand>"
                                   + "  <defaultBandName>userDefault</defaultBandName>"
                                   + "  <targetBands>"
                                   + "    <band>u</band>"
                                   + "    <band>v</band>"
                                   + "    <band>w</band>"
                                   + "  </targetBands>"
                                   + "</parameters>";
        final AnnotatedPojo value = new AnnotatedPojo();
        value.targetBandName = "radiance_13";
        value.defaultBandName = "userDefault";
        value.targetBandNames = new String[]{"u", "v", "w"};
        final XppDom dom = new XppDom("parameters");
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testDomToEnumPojo() throws Exception {

        final String xml = ""
                           + "<parameters>"
                           + "  <character>KERMIT</character>"
                           + "</parameters>";
        final XppDom dom = createDom(xml);
        final EnumPojo value = new EnumPojo();
        assertNull(value.character);
        convertDomToValue(dom, value);
        assertEquals(EnumPojo.Muppet.KERMIT, value.character);
    }

    public void testEnumPojoToDom() throws ValidationException, ConversionException {

        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <character>MISS_PIGGY</character>"
                                   + "</parameters>";
        final EnumPojo value = new EnumPojo();
        value.character = DefaultDomConverterTest.EnumPojo.Muppet.MISS_PIGGY;
        final XppDom dom = new XppDom("parameters");
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testDomToComplexPojo() throws Exception {

        final String xmlString = ""
                                 + "<parameters>"
                                 + "  <simple>"
                                 + "    <intField>42</intField>"
                                 + "    <stringField>a string</stringField>"
                                 + "    <doubleArrayField>1.2, 4.5, -0.034</doubleArrayField>"
                                 + "  </simple>"
                                 + "  <annotatedPojo>"
                                 + "    <targetBand>result</targetBand>"
                                 + "    <targetBands>"
                                 + "      <band>a</band>"
                                 + "      <band>b</band>"
                                 + "    </targetBands>"
                                 + "  </annotatedPojo>"
                                 + "</parameters>";
        final XppDom dom = createDom(xmlString);
        final ComplexPojo value = new ComplexPojo();
        assertNull(value.simplePojo);
        assertNull(value.annotatedPojo);

        convertDomToValue(dom, value);

        assertEquals(42, value.simplePojo.intField);
        assertEquals("a string", value.simplePojo.stringField);
        assertNotNull(value.simplePojo.doubleArrayField);
        assertEquals(3, value.simplePojo.doubleArrayField.length);
        assertEquals(1.2, value.simplePojo.doubleArrayField[0], 1.0e-10);
        assertEquals(4.5, value.simplePojo.doubleArrayField[1], 1.0e-10);
        assertEquals(-0.034, value.simplePojo.doubleArrayField[2], 1.0e-10);

        assertEquals("result", value.annotatedPojo.targetBandName);
        assertEquals("defaultBand", value.annotatedPojo.defaultBandName);
        assertNotNull(value.annotatedPojo.targetBandNames);
        assertEquals(2, value.annotatedPojo.targetBandNames.length);
        assertEquals("a", value.annotatedPojo.targetBandNames[0]);
        assertEquals("b", value.annotatedPojo.targetBandNames[1]);

    }

    public void testComplexPojoToDom() throws ValidationException, ConversionException {

        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <simple>"
                                   + "    <intField>87</intField>"
                                   + "    <stringField>Test, test, test!</stringField>"
                                   + "    <doubleArrayField>0.5,1.0</doubleArrayField>"
                                   + "    <fileField/>"
                                   + "  </simple>"
                                   + "  <annotatedPojo>"
                                   + "    <targetBand>reflec_4</targetBand>"
                                   + "  <defaultBandName>userDefault</defaultBandName>"
                                   + "    <targetBands>"
                                   + "      <band>real</band>"
                                   + "      <band>imag</band>"
                                   + "    </targetBands>"
                                   + "  </annotatedPojo>"
                                   + "</parameters>";
        final ComplexPojo value = new ComplexPojo();
        value.simplePojo = new SimplePojo();
        value.simplePojo.intField = 87;
        value.simplePojo.stringField = "Test, test, test!";
        value.simplePojo.doubleArrayField = new double[]{0.5, 1.0};
        value.annotatedPojo = new AnnotatedPojo();
        value.annotatedPojo.targetBandName = "reflec_4";
        value.annotatedPojo.defaultBandName = "userDefault";
        value.annotatedPojo.targetBandNames = new String[]{"real", "imag"};
        final XppDom dom = new XppDom("parameters");
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testDomToInlinedArrayPojo() throws Exception {

        final String xml = ""
                           + "<parameters>"
                           + "  <endmember>"
                           + "    <name>Land</name>"
                           + "    <size>4</size>"
                           + "    <wavelengths>820,830,840,850</wavelengths>"
                           + "    <radiances>220,230,240,250</radiances>"
                           + "  </endmember>"
                           + "  <defaultEndmember>"      // note the order!
                           + "    <name>Fallback</name>"
                           + "    <size>4</size>"
                           + "    <wavelengths>820,830,840,850</wavelengths>"
                           + "    <radiances>420,430,440,450</radiances>"
                           + "  </defaultEndmember>"
                           + "  <endmember>"
                           + "    <name>Water</name>"
                           + "    <size>4</size>"
                           + "    <wavelengths>820,830,840,850</wavelengths>"
                           + "    <radiances>620,630,640,650</radiances>"
                           + "  </endmember>"
                           + "  <endmember>"
                           + "    <name>Cloud</name>"
                           + "    <size>4</size>"
                           + "    <wavelengths>820,830,840,850</wavelengths>"
                           + "    <radiances>920,930,940,950</radiances>"
                           + "  </endmember>"
                           + "</parameters>";

        final XppDom dom = createDom(xml);
        final InlinedArrayPojo value = new InlinedArrayPojo();
        assertNull(value.defaultEndmember);
        assertNull(value.endmembers);

        convertDomToValue(dom, value);

        assertNotNull(value.defaultEndmember);
        assertEquals("Fallback", value.defaultEndmember.name);
        assertEquals(4, value.defaultEndmember.size);
        assertNotNull(value.defaultEndmember.wavelengths);
        assertNotNull(value.defaultEndmember.radiances);

        assertNotNull(value.endmembers);
        assertEquals(3, value.endmembers.length);

        assertEquals("Land", value.endmembers[0].name);
        assertEquals(4, value.endmembers[0].size);
        assertNotNull(value.endmembers[0].wavelengths);
        assertNotNull(value.endmembers[0].radiances);

        assertEquals("Water", value.endmembers[1].name);
        assertEquals(4, value.endmembers[1].size);
        assertNotNull(value.endmembers[1].wavelengths);
        assertNotNull(value.endmembers[1].radiances);

        assertEquals("Cloud", value.endmembers[2].name);
        assertEquals(4, value.endmembers[2].size);
        assertNotNull(value.endmembers[2].wavelengths);
        assertNotNull(value.endmembers[2].radiances);
    }

    public void testInlinedArrayPojoToDom() throws ValidationException, ConversionException {

        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <defaultEndmember>"
                                   + "    <name>Fallback</name>"
                                   + "    <size>4</size>"
                                   + "    <wavelengths>820.0,830.0,840.0,850.0</wavelengths>"
                                   + "    <radiances>420.0,430.0,440.0,450.0</radiances>"
                                   + "  </defaultEndmember>"
                                   + "  <endmember>"
                                   + "    <name>Land</name>"
                                   + "    <size>4</size>"
                                   + "    <wavelengths>820.0,830.0,840.0,850.0</wavelengths>"
                                   + "    <radiances>220.0,230.0,240.0,250.0</radiances>"
                                   + "  </endmember>"
                                   + "  <endmember>"
                                   + "    <name>Water</name>"
                                   + "    <size>4</size>"
                                   + "    <wavelengths>820.0,830.0,840.0,850.0</wavelengths>"
                                   + "    <radiances>620.0,630.0,640.0,650.0</radiances>"
                                   + "  </endmember>"
                                   + "  <endmember>"
                                   + "    <name>Cloud</name>"
                                   + "    <size>4</size>"
                                   + "    <wavelengths>820.0,830.0,840.0,850.0</wavelengths>"
                                   + "    <radiances>920.0,930.0,940.0,950.0</radiances>"
                                   + "  </endmember>"
                                   + "</parameters>";

        final InlinedArrayPojo value = new InlinedArrayPojo();
        value.defaultEndmember = new Endmember("Fallback", new double[]{820, 830, 840, 850},
                                               new double[]{420, 430, 440, 450});
        value.endmembers = new Endmember[3];
        value.endmembers[0] = new Endmember("Land", new double[]{820, 830, 840, 850}, new double[]{220, 230, 240, 250});
        value.endmembers[1] = new Endmember("Water", new double[]{820, 830, 840, 850},
                                            new double[]{620, 630, 640, 650});
        value.endmembers[2] = new Endmember("Cloud", new double[]{820, 830, 840, 850},
                                            new double[]{920, 930, 940, 950});

        final XppDom dom = new XppDom("parameters");
        convertValueToDom(value, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testInterfaceFieldsPojoToDom() throws ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <shape1 class=\"java.awt.Rectangle\" >"
                                   + "    <x>10</x>"
                                   + "    <y>10</y>"
                                   + "    <width>20</width>"
                                   + "    <height>25</height>"
                                   + "  </shape1>"
                                   + "  <shape2 class=\"java.awt.geom.Line2D$Float\" >"
                                   + "    <x1>0.0</x1>"
                                   + "    <y1>10.3</y1>"
                                   + "    <x2>15.7</x2>"
                                   + "    <y2>34.6</y2>"
                                   + "  </shape2>"
                                   + "</parameters>";
        final InterfaceFieldsPojo interfacePojo = new InterfaceFieldsPojo(new Rectangle(10, 10, 20, 25),
                                                                          new Line2D.Float(0.0f, 10.3f, 15.7f, 34.6f));
        final XppDom dom = new XppDom("parameters");
        convertValueToDom(interfacePojo, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void testDomToInterfaceFieldsPojo() throws Exception {
        final String xml = ""
                           + "<parameters>"
                           + "  <shape1 class=\"java.awt.Rectangle\" >"
                           + "    <x>10</x>"
                           + "    <y>10</y>"
                           + "    <width>20</width>"
                           + "    <height>25</height>"
                           + "  </shape1>"
                           + "  <shape2 class=\"java.awt.geom.Line2D$Float\" >"
                           + "    <x1>0.0</x1>"
                           + "    <y1>10.3</y1>"
                           + "    <x2>15.7</x2>"
                           + "    <y2>34.6</y2>"
                           + "  </shape2>"
                           + "</parameters>";

        final XppDom dom = createDom(xml);
        final InterfaceFieldsPojo value = new InterfaceFieldsPojo(null, null);
        assertNull(value.shape1);
        assertNull(value.shape2);

        convertDomToValue(dom, value);

        assertNotNull(value.shape1);
        assertNotNull(value.shape1 instanceof Rectangle);
        final Rectangle shape1 = (Rectangle) value.shape1;
        assertEquals(10, shape1.x);
        assertEquals(10, shape1.x);
        assertEquals(20, shape1.width);
        assertEquals(25, shape1.height);

        assertNotNull(value.shape2);
        assertNotNull(value.shape2 instanceof Line2D.Float);
        final Line2D.Float shape2 = (Line2D.Float) value.shape2;
        assertEquals(0.0f, shape2.x1);
        assertEquals(10.3f, shape2.y1);
        assertEquals(15.7f, shape2.x2);
        assertEquals(34.6f, shape2.y2);
    }

    public void doNotTestMapFieldPojoToDom() throws ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <map class=\"java.util.HashMap\" >"
                                   + "    <entry>"
                                   + "      <key class=\"java.lang.String\">Bibo</key>"
                                   + "      <value class=\"java.awt.Rectangle\">"
                                   + "        <x>10</x>"
                                   + "        <y>10</y>"
                                   + "        <width>20</width>"
                                   + "        <height>25</height>"
                                   + "      </value>"
                                   + "    </entry>"
                                   + "    <entry>"
                                   + "      <key class=\"java.lang.Integer\">12345</key>"
                                   + "      <value class=\"java.awt.Color\">12,40,123</value>"
                                   + "    </entry>"
                                   + "  </map>"
                                   + "</parameters>";

        Map<Object, Object> map = new HashMap<Object, Object>();
        map.put("Bibo", new Rectangle(10, 10, 20, 25));
        map.put(12345, new Color(12, 40, 123));
        final MapFieldPojo mapFieldPojo = new MapFieldPojo(map);
        final XppDom dom = new XppDom("parameters");
        convertValueToDom(mapFieldPojo, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void doNotTestDomToMapFieldPojo() throws Exception {
        final String xml = ""
                           + "<parameters>"
                           + "  <map class=\"java.util.HashMap\" >"
                           + "    <entry>"
                           + "      <key class=\"java.lang.String\">Bibo</key>"
                           + "      <value class=\"java.awt.Rectangle\">"
                           + "        <x>10</x>"
                           + "        <y>10</y>"
                           + "        <width>20</width>"
                           + "        <height>25</height>"
                           + "      </value>"
                           + "    </entry>"
                           + "    <entry>"
                           + "      <key class=\"java.lang.Integer\">12345</key>"
                           + "      <value class=\"java.awt.Color\">12,40,123</value>"
                           + "    </entry>"
                           + "  </map>"
                           + "</parameters>";

        final XppDom dom = createDom(xml);
        final MapFieldPojo mapFieldPojo = new MapFieldPojo(null);
        convertDomToValue(dom, mapFieldPojo);

        final Map<?, ?> map = mapFieldPojo.map;
        assertTrue(map instanceof HashMap);
        assertEquals(2, map.size());
    }

    public void doNotTestCollectionFieldPojoToDom() throws ConversionException {
        final String expectedXml = ""
                                   + "<parameters>"
                                   + "  <collection class=\"java.util.Stack\">"
                                   + "    <item class=\"java.awt.Rectangle\">"
                                   + "      <x>0</x>"
                                   + "      <y>0</y>"
                                   + "      <width>10</width>"
                                   + "      <height>10</height>"
                                   + "    </item>"
                                   + "    <item class=\"java.awt.geom.Line2D$Double\">"
                                   + "      <x1>0.0</x1>"
                                   + "      <y1>0.0</y1>"
                                   + "      <x2>10.0</x2>"
                                   + "      <y2>10.0</y2>"
                                   + "    </item>"
                                   + "    <item class=\"java.awt.geom.Arc2D$Double\">"
                                   + "      <type>1</type>"
                                   + "      <x>0.0</x>"
                                   + "      <y>0.0</y>"
                                   + "      <width>10.0</width>"
                                   + "      <height>10.0</height>"
                                   + "      <start>2.0</start>"
                                   + "      <extent>3.0</extent>"
                                   + "    </item>"
                                   + "  </collection>"
                                   + "</parameters>";
        Collection<Shape> stack = new Stack<Shape>();
        stack.add(new Rectangle(0, 0, 10, 10));
        stack.add(new Line2D.Double(0, 0, 10, 10));
        stack.add(new Arc2D.Double(0, 0, 10, 10, 2, 3, Arc2D.CHORD));
        final CollectionFieldPojo collectionFieldPojo = new CollectionFieldPojo(stack);
        final XppDom dom = new XppDom("parameters");
        convertValueToDom(collectionFieldPojo, dom);
        assertDomEquals(createDom(expectedXml), dom);
    }

    public void doNotTestDomToCollectionFieldPojo() throws Exception {
        final String xml = ""
                           + "<parameters>"
                           + "  <collection class=\"java.util.Stack\">"
                           + "    <item class=\"java.awt.Rectangle\">"
                           + "      <x>0</x>"
                           + "      <y>0</y>"
                           + "      <width>10</width>"
                           + "      <height>10</height>"
                           + "    </item>"
                           + "    <item class=\"java.awt.geom.Line2D$Double\">"
                           + "      <x1>0.0</x1>"
                           + "      <y1>0.0</y1>"
                           + "      <x2>10.0</x2>"
                           + "      <y2>10.0</y2>"
                           + "    </item>"
                           + "    <item class=\"java.awt.geom.Arc2D$Double\">"
                           + "      <type>1</type>"
                           + "      <x>0.0</x>"
                           + "      <y>0.0</y>"
                           + "      <width>10.0</width>"
                           + "      <height>10.0</height>"
                           + "      <start>2.0</start>"
                           + "      <extent>3.0</extent>"
                           + "    </item>"
                           + "  </collection>"
                           + "</parameters>";

        final XppDom dom = createDom(xml);
        final CollectionFieldPojo collectionFieldPojo = new CollectionFieldPojo(null);
        convertDomToValue(dom, collectionFieldPojo);

        assertTrue(collectionFieldPojo.collection instanceof Stack);
        Stack stack = (Stack) collectionFieldPojo.collection;
        assertEquals(3, stack.size());
        assertTrue(stack.get(0) instanceof Rectangle);
        assertTrue(stack.get(1) instanceof Line2D.Double);
        assertTrue(stack.get(2) instanceof java.awt.geom.Arc2D.Double);
    }

    private XppDom createDom(String xml) {
        XppDomWriter domWriter = new XppDomWriter();
        new HierarchicalStreamCopier().copy(new XppReader(new StringReader(xml)), domWriter);
        return domWriter.getConfiguration();
    }

    public static void convertValueToDom(Object value, XppDom parentElement) throws ConversionException {
        DefaultDomConverter domConverter = new DefaultDomConverter(value.getClass(), VALUE_DESCRIPTOR_FACTORY);
        XppDomElement domElement = new XppDomElement(parentElement);
        domConverter.convertValueToDom(value, domElement);
    }


    public static void convertDomToValue(XppDom parentElement, Object value) throws Exception {
        DefaultDomConverter domConverter = new DefaultDomConverter(value.getClass(), VALUE_DESCRIPTOR_FACTORY);
        XppDomElement domElement = new XppDomElement(parentElement);
        domConverter.convertDomToValue(domElement, value);
    }

    public static class SimplePojo {

        File fileField;
        String stringField;
        int intField;
        double[] doubleArrayField;
    }

    public static class AnnotatedPojo {

        @X(alias = "targetBand")
        String targetBandName;

        @X(defaultValue = "defaultBand")
        String defaultBandName;

        @X(alias = "targetBands", componentAlias = "band")
        String[] targetBandNames;
    }

    public static class ComplexPojo {

        @X(alias = "simple")
        SimplePojo simplePojo;
        AnnotatedPojo annotatedPojo;
    }

    public static class MemberDomConverter implements DomConverter {

        private final DefaultDomConverter memberConverter = new DefaultDomConverter(Member.class);

        @Override
        public Class<?> getValueType() {
            return Member.class;
        }

        @Override
        public Object convertDomToValue(DomElement parentElement, Object value) throws ConversionException,
                                                                                       ValidationException {
            DomElement[] children = parentElement.getChildren("member");
            Member[] members = new Member[children.length];
            for (int i = 0; i < children.length; i++) {
                members[i] = (Member) memberConverter.convertDomToValue(children[i], null);
            }
            return members;
        }

        @Override
        public void convertValueToDom(Object value, DomElement parentElement) throws ConversionException {
            Member[] members = (Member[]) value;
            for (Member member : members) {
                DomElement aggregator = parentElement.createChild("member");
                memberConverter.convertValueToDom(member, aggregator);
            }
        }
    }

    public static class ArrayWithDomConverterPojo {

        @X(domConverter = MemberDomConverter.class)
        Member[] allies;
    }

    public static class ArrayPojo {

        Member prince;
        Member[] allies;

    }

    public static class Member {

        private final String name;

        public Member(String name) {
            this.name = name;
        }
    }

    public static class EnumPojo {

        private enum Muppet {
            KERMIT,
            ANIMAL,
            MISS_PIGGY
        }

        Muppet character;
    }

    public static class InlinedArrayPojo {

        Endmember defaultEndmember;

        @X(componentAlias = "endmember", inlined = true)
        Endmember[] endmembers;
    }

    public static class InterfaceFieldsPojo {

        Shape shape1;
        Shape shape2;

        public InterfaceFieldsPojo(Shape shape1, Shape shape2) {
            this.shape1 = shape1;
            this.shape2 = shape2;
        }
    }

    public static class MapFieldPojo {

        Map<?, ?> map;

        public MapFieldPojo(Map<?, ?> map) {
            this.map = map;
        }
    }

    public static class CollectionFieldPojo {

        Collection collection;

        public CollectionFieldPojo(Collection collection) {
            this.collection = collection;
        }
    }

    public static class WeirdPojo {

        final Weird finalWeird = new Weird();
        Weird weird;
        String name;
        transient Date timeStamp = new Date();
    }

    public static class Weird {

        static final Weird staticFinalWeird = new Weird();
        static Weird staticWeird = new Weird();
    }

    @Retention(value = RetentionPolicy.RUNTIME)
    @interface X {

        String alias() default "";

        String componentAlias() default "";

        boolean inlined() default false;

        String defaultValue() default "";

        Class<? extends DomConverter> domConverter() default DomConverter.class;
    }


    public static class Endmember {

        public Endmember() {
        }

        public Endmember(String name, double[] wavelengths, double[] radiances) {
            this.name = name;
            this.size = wavelengths.length;
            this.wavelengths = wavelengths;
            this.radiances = radiances;
        }

        String name;
        int size;
        double[] wavelengths;
        double[] radiances;
    }

    public static void assertDomEquals(XppDom expected, XppDom actual) {
        assertDomEquals("", expected, actual);
    }

    public static void assertDomEquals(String message, XppDom expected, XppDom actual) {
        message = message + "/" + expected.getName();
        assertEquals(message, expected.getName(), actual.getName());
        assertEquals(message, expected.getChildCount(), actual.getChildCount());
        if (expected.getChildCount() == 0) {
            assertEquals(message, expected.getValue(), actual.getValue());
        } else {
            assertTrue(message, expected.getValue() == null || expected.getValue().trim().isEmpty());
            assertTrue(message, actual.getValue() == null || actual.getValue().trim().isEmpty());
        }
        final String[] attributeNames = expected.getAttributeNames();
        for (String expectedAttrName : attributeNames) {
            final String expectedAttrValue = expected.getAttribute(expectedAttrName);
            final String actualAttrValue = actual.getAttribute(expectedAttrName);
            assertEquals(message, expectedAttrValue, actualAttrValue);
        }
        final XppDom[] expectedChildren = expected.getChildren();
        final XppDom[] actualChildren = actual.getChildren();
        for (XppDom expectedChild : expectedChildren) {
            boolean success = false;
            for (XppDom actualChild : actualChildren) {
                try {
                    assertDomEquals(message, expectedChild, actualChild);
                    success = true;
                    break;
                } catch (AssertionFailedError e) {
                    // try next one
                }
            }
            if (!success) {
                final XppDom actualChild = actual.getChild(expectedChild.getName());
                assertNotNull(message, actualChild);
                assertDomEquals(message, expectedChild, actualChild);
            }

        }
    }
}
